from unittest.mock import MagicMock, patch, call
from rebotics_sdk.hook.views import WebhookHandlerViewSet
from django.dispatch import receiver
from rest_framework import status
from rest_framework.test import APITestCase, APIRequestFactory

from rebotics_sdk.hook.signals import PROCESSING_FINISH_SUCCESS, EVENT_SIGNALS


class SignalEmitterTestCase(APITestCase):
    def test_success(self):
        mock = MagicMock()
        mock.return_value = None
        receiver(PROCESSING_FINISH_SUCCESS)(mock)

        factory = APIRequestFactory()
        request = factory.post(
            "/webhook/testretailer/",
            data={
                "event": "processing.finish",
                "payload": {
                    "id": 123,
                    "status": "done",
                }
            },
            format='json'
        )
        with patch("rebotics_sdk.hook.signals.SignalEmitter.send_signal") as signal_mock:
            view = WebhookHandlerViewSet.as_view()
            response = view(request, retailer_code="testretailer")
            self.assertEqual(response.status_code, status.HTTP_202_ACCEPTED)
            signal_mock.assert_called_once()

    def test_success_called_with_check(self):
        mock = MagicMock()
        mock.return_value = None
        receiver(PROCESSING_FINISH_SUCCESS)(mock)

        factory = APIRequestFactory()
        request = factory.post(
            "/webhook/testretailer/",
            data={
                "event": "processing.finish",
                "payload": {
                    "id": 123,
                    "status": "done",
                }
            },
            format='json'
        )
        with patch("rebotics_sdk.hook.signals.SignalEmitter") as signal_mock_2:
            view = WebhookHandlerViewSet.as_view()
            response = view(request, retailer_code="testretailer")
            self.assertEqual(response.status_code, status.HTTP_202_ACCEPTED)
            signal_mock_2.assert_called_with(
                retailer="testretailer",
                event="processing.finish",
                payload={
                    "id": 123,
                    "status": "done",
                }
            )

    def test_no_fields(self):
        factory = APIRequestFactory()
        request = factory.post(
            "/webhook/testretailer/",
            data={},
            format='json'
        )

        view = WebhookHandlerViewSet.as_view()
        response = view(request, retailer_code="testretailer")
        error_data = {
            "event": ["This field is required."],
            "payload": ["This field is required."]
        }
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.data, error_data)

    def test_different_event(self):
        factory = APIRequestFactory()
        request = factory.post(
            "/webhook/testretailer/",
            data={
                "event": "hello",
                "payload": {
                    "id": 123,
                    "status": "done",
                }
            },
            format='json'
        )

        view = WebhookHandlerViewSet.as_view()
        response = view(request, retailer_code="testretailer")
        error_data = {
            "event": ["\"hello\" is not a valid choice."]
        }
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.data, error_data)

    def test_no_status(self):
        factory = APIRequestFactory()
        request = factory.post(
            "/webhook/testretailer/",
            data={
                "event": "processing.finish",
                "payload": {
                    "id": 123
                }
            },
            format='json'
        )

        view = WebhookHandlerViewSet.as_view()
        response = view(request, retailer_code="testretailer")
        error_data = {
            "payload":
                {
                    "status": ["This field is required."]
                }
        }
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.data, error_data)



import unittest
import random
import string
from urllib.parse import urljoin, urlparse, unquote

from rebotics_sdk.providers import FVMProvider


class GenerateTokenTestCase(unittest.TestCase):
    def test_success(self):
        codename = 'retailer_x'
        secret_key = 'secret_key_x'
        file_url = 'file_url'
        path = '/api/previews/'
        barcode = '21152'
        scheme = 'https'
        netloc = 'r3dev-fvm.rebotics.net'
        allowed_chars = ''.join((string.ascii_letters, string.digits))
        unique_id = ''.join(random.choice(allowed_chars) for _ in range(44))
        fvm_provider = FVMProvider(f'{scheme}://{netloc}/')
        fvm_provider.set_retailer_identifier(codename, secret_key)
        token = fvm_provider.generate_token(file_url)
        self.assertNotEqual(token, unique_id)
        permalink = fvm_provider.get_preview_permalink(barcode)
        parsed_permalink = urlparse(permalink)
        self.assertEqual(permalink, unquote(permalink))
        self.assertEqual(parsed_permalink.scheme, scheme)
        self.assertEqual(parsed_permalink.netloc, netloc)
        self.assertEqual(parsed_permalink.path, f'{path}{barcode}/')
        self.assertEqual(parsed_permalink.query.split("codename=")[1].split("&")[0], codename)

    def test_raise_exception(self):
        fvm_provider = FVMProvider('host')
        with self.assertRaises(AttributeError):
            fvm_provider.generate_token('file_url')



from .base import ReboticsBaseProvider, remote_service


class HawkeyeProvider(ReboticsBaseProvider):

    @remote_service('/api-token-auth/', raw=True)
    def token_auth(self, username, password, **kwargs):
        response = self.session.post(data={
            'username': username,
            'password': password
        })
        self.set_token(response.json()['token'])
        return response

    @remote_service('/api/v1/camera/heartbeats/')
    def save_camera_heartbeat(self, shelf_camera, battery_status, free_storage, used_storage):
        return self.session.post(json={
            "shelf_camera": shelf_camera,
            "battery_status": battery_status,
            "free_storage": free_storage,
            "used_storage": used_storage
        })

    @remote_service('/api/v1/camera/camera-actions/')
    def save_camera_action(self, action_type, status_type, payload, shelf_camera):
        return self.session.post(json={
            "action_type": action_type,
            "status_type": status_type,
            "payload": payload,
            "shelf_camera": shelf_camera
        })

    @remote_service('/api/v1/camera/camera-actions/')
    def get_camera_actions(self):
        return self.session.get()

    @remote_service('/api/v1/camera/retailer/')
    def save_retailer(self, codename, token):
        return self.session.post(json={
            "retailer": codename,
            "token": token
        })

    @remote_service('/api/v1/camera/fixtures/')
    def save_fixture(self, retailer, store_id, aisle, section):
        return self.session.post(
            json={
                "store_id": store_id,
                "aisle": aisle,
                "section": section,
                "retailer": retailer
            }
        )

    @remote_service('/api/v1/camera/fixtures/{id}')
    def delete_fixture(self, pk):
        return self.session.delete(id=pk)

    @remote_service('/api/v1/camera/fixtures/')
    def get_fixtures(self):
        return self.session.get()

    @remote_service('/api/v1/camera/shelf-cameras/')
    def create_shelf_camera(self, camera_id, added_by, fixture=None):
        data = {
            "camera_id": camera_id,
            "added_by": added_by,
        }
        if fixture is not None:
            data["fixture"] = fixture
        return self.session.post(json=data)

    @remote_service('/api/v1/camera/shelf-cameras/')
    def get_shelf_cameras(self):
        return self.session.get()

    @remote_service('/api/v1/fetcher/{camera_id}/')
    def save_capture(self, camera, file_key, bucket_name):
        return self.session.post(
            camera_id=camera,
            json={
                "file_key": file_key,
                "bucket_name": bucket_name
            })

    @remote_service('/api/v1/camera/mac-address/search/')
    def search_camera_by_mac(self, mac_address):
        assert ":" in mac_address, "Mac address should be separated by : (colon)"
        return self.session.post(json={
            'mac_address': mac_address
        })

    @remote_service('/api/v1/camera/mac-address/')
    def create_mac_address_link(self, mac_address, vendor_secret, rebotics_secret):
        assert ":" in mac_address, "Mac address should be separated by : (colon)"
        return self.session.post(json={
            'mac_address': mac_address,
            'vendor_secret': vendor_secret,
            'rebotics_secret': rebotics_secret,
        })

    @remote_service('/api/v1/camera/mac-address/{id}/')
    def update_mac_address_link(self, id_, mac_address, vendor_secret, rebotics_secret):
        assert ":" in mac_address, "Mac address should be separated by : (colon)"
        assert isinstance(id_, int), "Link ID should be integer, use search by mac-address to determine Link ID"
        return self.session.patch(id=id_, json={
            'mac_address': mac_address,
            'vendor_secret': vendor_secret,
            'rebotics_secret': rebotics_secret,
        })

    @remote_service('/api/v1/camera/mac-address/{id}/', raw=True)
    def delete_mac_address_link(self, id_):
        response = self.session.delete(id=id_)
        response.raise_for_status()



from rest_framework import serializers

from rebotics_sdk.hook.signals import EVENTS


class WebhookPayloadSerializer(serializers.Serializer):
    id = serializers.IntegerField()
    status = serializers.CharField()


class WebhookDataSerializer(serializers.Serializer):
    event = serializers.ChoiceField(choices=EVENTS)
    payload = WebhookPayloadSerializer()



import shutil
import tempfile
from pathlib import Path
import requests_mock

from rebotics_sdk.providers import FVMProvider

RESOURCES_FOLDER = Path(__file__).parent / 'resources'

DATA_VIRTUAL = {
        'filename': 'qwerty.csv',
        'id': 1,
        'destination': {
            'url': 'http://minio:9005/demo-bucket',
            'fields': {
                'key': 'virtual_uploads/1/qwerty.csv',
                'AWSAccessKeyId': 'access_key',
                'policy': 'eyJleHBpcmF0aW9uIjogIjIwMjEtMTAtMDdUMDc6NDI6MDBaIiwgImNvbm'
                          'RpdGlvbnMiOiBbeyJidWNrZXQiOiAiZGVtby1idWNrZXQifSwgeyJrZXkiOi'
                          'AidmlydHVhbF91cGxvYWRzLzEvcXdlcnR5LmNzdiJ9XX0=',
                'signature': 'zF2lZqmQPeMwEod1+pMOHjObYVE='
            }
        }
    }
DATA_FINISH = {
    'id': 1,
    'file': 'http://minio:9005/demo-bucket/example_TEi7koV.jpg?'
            'AWSAccessKeyId=access_key&Signature=3J01a4yIbiVeUqCppBD4AKGkY%2FE%3D&Expires=1633598295',
    'file_key': 'example_TEi7koV.jpg'

}


def test_create_virtual_upload_success(runner, mocker):
    from unittest.mock import Mock
    process_role_mock = mocker.patch("rebotics_sdk.cli.utils.process_role")
    process_role_mock.return_value = None
    provider_mock = Mock(spec=FVMProvider)
    mocker.patch('rebotics_sdk.cli.utils.ReboticsCLIContext.provider', new=provider_mock)
    provider_mock.create_virtual_upload.side_effect = lambda *args: DATA_VIRTUAL
    provider_mock.finish.side_effect = lambda *args: DATA_FINISH
    from rebotics_sdk.cli.fvm import api
    with requests_mock.Mocker() as m:
        m.post('http://minio:9005/demo-bucket', status_code=200)
        with tempfile.TemporaryDirectory() as base_dir:
            base_dir_path = Path(base_dir)
            filename = base_dir_path.absolute() / 'qwerty.csv'
            shutil.copy(RESOURCES_FOLDER / 'qwerty.csv', filename)

            result = runner.invoke(
                api,
                [
                    'file',
                    '-f', str(filename),
                ]
            )
        provider_mock.create_virtual_upload.assert_called_once_with(
            str(filename)
        )
    assert result.exit_code == 0
    assert not result.exception
    assert "Successfully finished uploading" in result.output.strip()


def test_create_virtual_upload_missing_option(runner, mocker):
    from unittest.mock import Mock
    process_role_mock = mocker.patch("rebotics_sdk.cli.utils.process_role")
    process_role_mock.return_value = None
    provider_mock = Mock(spec=FVMProvider)
    mocker.patch('rebotics_sdk.cli.utils.ReboticsCLIContext.provider', new=provider_mock)
    provider_mock.create_virtual_upload.side_effect = lambda *args: DATA_VIRTUAL
    provider_mock.finish.side_effect = lambda *args: DATA_FINISH
    from rebotics_sdk.cli.fvm import api
    result = runner.invoke(
        api,
        [
            'file'
        ]
    )
    assert result.exit_code == 2
    assert result.exception
    assert "Missing option" in result.output


def test_fv_api_create_virtual_upload_no_file(runner, mocker):
    from unittest.mock import Mock
    process_role_mock = mocker.patch("rebotics_sdk.cli.utils.process_role")
    process_role_mock.return_value = None
    provider_mock = Mock(spec=FVMProvider)
    mocker.patch('rebotics_sdk.cli.utils.ReboticsCLIContext.provider', new=provider_mock)
    provider_mock.create_virtual_upload.side_effect = lambda *args: DATA_VIRTUAL
    provider_mock.finish.side_effect = lambda *args: DATA_FINISH
    from rebotics_sdk.cli.fvm import api
    result = runner.invoke(
        api,
        [
            'file',
            '-f', 'lol',
        ]
    )
    assert result.exit_code == 1
    assert result.exception



import logging
import os

import click

from .common import shell, roles, configure, set_token
from .utils import ReboticsCLIContext, pass_rebotics_context, app_dir, process_role, read_saved_role
from ..providers import RetailerProvider

logger = logging.getLogger(__name__)


@click.group()
@click.option('-f', '--format', default='table', type=click.Choice(['table', 'id', 'json']))
@click.option('-v', '--verbose', is_flag=True, help='Enables verbose mode')
@click.option('-c', '--config', type=click.Path(), default='fvm.json', help="Specify what config.json to use")
@click.option('-r', '--role', default=lambda: read_saved_role('retailer'), help="Key to specify what retailer to use")
@click.version_option()
@click.pass_context
def api(ctx, format, verbose, config, role):
    """
    Retailer CLI tool to communicate with retailer API
    """
    process_role(ctx, role, 'fvm')
    ctx.obj = ReboticsCLIContext(
        role,
        format,
        verbose,
        os.path.join(app_dir, config),
        provider_class=RetailerProvider
    )


@api.command()
@pass_rebotics_context
def version(ctx):
    """Show retailer backend version"""
    ctx.format_result(ctx.provider.version(), 100)


@api.command()
@click.option('-t', '--input_type')
@click.option('-s', '--store', type=click.INT)
@click.argument('files', nargs=-1, required=True, type=click.File('rb'))
@pass_rebotics_context
def upload_files(ctx, input_type, store, files):
    """
    Upload processing files to the retailer backend, that can be used as processing action inputs
    """
    file_ids = []
    for f_ in files:
        response = ctx.provider.processing_upload(
            store, f_, input_type
        )
        file_ids.append(response['id'])

        if ctx.verbose:
            click.echo(response)  # redirecting this output to stderr
    click.echo(' '.join(map(str, file_ids)))


@api.group(name='fv')
def feature_vectors():
    """
    Feature vector flow
    """
    pass


@feature_vectors.command('list')
@pass_rebotics_context
def feature_vectors_list(ctx):
    fve = ctx.provider.feature_vectors_export()
    data = fve.get()
    if data:
        ctx.format_result(data)


@feature_vectors.command('export')
@click.option('-b', '--batch-size', type=click.INT, default=50000)
@click.option('-s', '--source-model', default='')
@click.option('-r', '--result-model', default='previews-backup')
@pass_rebotics_context
def feature_vectors_export(ctx, source_model, result_model, batch_size):
    fve = ctx.provider.feature_vectors_export()
    result = fve.export(source_model=source_model, result_model=result_model, batch_size=batch_size)
    ctx.format_result(result)


api.add_command(shell, 'shell')
api.add_command(roles, 'roles')
api.add_command(configure, 'configure')
api.add_command(set_token, 'set_token')



import hashlib
import os
import pathlib
import shlex
import subprocess as s
import uuid
from datetime import datetime
from pprint import pformat

import chardet
import click
import six
import yaml

from rebotics_sdk.cli.renderers import format_full_table
from rebotics_sdk.cli.utils import pass_rebotics_context
from rebotics_sdk.providers import ProviderHTTPClientException


def mean(arr):
    return sum(arr) / len(arr)


@click.group()
@click.version_option()
def main():
    """Collection of scripts for rebotics"""
    pass


@main.command()
@click.argument('container')
@click.argument('volume')
def fix_pycharm(container, volume):
    with s.Popen(shlex.split("docker ps -a --format '{{.Names}}'"), stdout=s.PIPE) as proc:
        containers = proc.stdout.read().decode('utf-8').split('\n')
        assert container in containers, 'Container should exist'

    # save helpers to the tmp folder
    tmp_folder = '/tmp/helper_folder_%s' % uuid.uuid4()
    os.system('docker cp {container}:/opt/.pycharm_helpers {tmp_folder}'.format(
        container=container,
        tmp_folder=tmp_folder
    ))

    # assign new container with attached volume
    helper_random_name = 'helper_%s' % uuid.uuid4()
    os.system('docker run -v {target_volume}:/opt/.pycharm_helpers --name {name} busybox true'.format(
        target_volume=volume,
        name=helper_random_name,
    ))

    os.system('docker cp {tmp_folder} {name}:/opt/.pycharm_helpers'.format(
        tmp_folder=tmp_folder,
        name=helper_random_name
    ))

    # Cleanup
    os.system('docker rm -f {helper_random_name}'.format(helper_random_name=helper_random_name))
    os.system('rm -rf {tmp_folder}'.format(tmp_folder=tmp_folder))
    click.echo('Finished')


@main.command()
@click.option("-p", "--pycharm_folder", help="pycharm folder from where to start", type=click.Path())
def fix_pycharm_clean(pycharm_folder):
    """
    This command fixes all pycharm related volumes and merges the helpers folder to it.
    This is in case when you update pycharm and run with debugger, but it yields error:
    python: can't open file '/opt/.pycharm_helpers/pydev/pydevd.py': [Errno 2] No such file or directory
    python: can't open file '/opt/.pycharm_helpers/generator3/__main__.py': [Errno 2] No such file or directory
    """

    def find_pycharm_helpers_folders(base_folder):
        for dirpath, dirnames, filenames in os.walk(os.path.expanduser(base_folder)):
            if 'PyCharm' in dirpath:
                if '/helpers' in dirpath and '/helpers/' not in dirpath:  # get only the top folder
                    if 'pydev' in dirnames:
                        yield dirpath

    helper_folder = ''
    if pycharm_folder is not None:
        base_folder = pathlib.Path(pycharm_folder)
    else:
        base_folder = "~"

    for path in find_pycharm_helpers_folders(base_folder):
        helper_folder = path
        break
    click.echo('Using helper folder: {helper_folder}'.format(
        helper_folder=helper_folder
    ))

    with s.Popen(shlex.split("docker volume ls --format '{{ .Name }}'"), stdout=s.PIPE) as proc:
        volumes = proc.stdout.read().decode('utf-8').split('\n')

    for volume in volumes:
        if 'helpers' not in volume:
            continue
        click.echo('Fixing pycharm_helpers for {volume}'.format(volume=volume))
        # assign new container with attached volume
        container = 'helper_%s' % uuid.uuid4()
        os.system('docker run -v {volume}:/opt/.pycharm_helpers --name {container} busybox true'.format(
            volume=volume,
            container=container
        ))
        os.system('docker cp {helper_folder}/. {container}:/opt/.pycharm_helpers/'.format(
            helper_folder=helper_folder,
            container=container
        ))

        # Cleanup
        os.system('docker rm -f {container}'.format(container=container))
    click.echo('Finished')


@main.command()
def install_bash_completion():
    path = os.path.expanduser('~/.bashrc')

    with open(path, 'a') as bashrc:
        commands = [
            '# Rebotics scripts autocomplete scripts',
            '# After you uninstall rebotics-scripts, please consider to delete this too',
            'eval "$(_RETAILER_COMPLETE=source retailer)"',
            'eval "$(_ADMIN_COMPLETE=source admin)"',
            'eval "$(_DATASET_COMPLETE=source dataset)"',
            'eval "$(_REBOTICS_COMPLETE=source rebotics)"',
            'eval "$(_CAMERA_MANAGER_COMPLETE=source camera_manager)"',
        ]
        for command in commands:
            bashrc.write(command + "\n")

    click.edit(filename=path)


@main.command()
@click.argument('filepath', type=click.Path(exists=True, dir_okay=False))
@click.option('-e', '--encoding')
@click.option('-t', '--target', type=click.Path(exists=False))
def encode_utf8(filepath, encoding, target):
    if not encoding:
        with open(filepath, 'rb') as f_:
            possible_encoding_detected = chardet.detect(f_.read())
            encoding = possible_encoding_detected['encoding']
    try:
        import pandas as pd
    except ImportError:
        raise click.ClickException('You need to have pandas installed to use this feature')

    try:
        df = pd.read_csv(filepath, encoding=encoding, dtype=object)
    except Exception as exc:
        raise click.ClickException('Can not read the file. Error: %s' % exc)

    if not target:
        full_name, extension = filepath.rsplit('.', 1)
        target = '{}_utf8.{}'.format(full_name, extension)

    try:
        df.to_csv(target, encoding='utf-8')
    except Exception as exc:
        raise click.ClickException('Failed to save the csv file, %s' % exc)
    click.echo('File saved at {}'.format(target))


@click.command()
@pass_rebotics_context
def shell(ctx):
    """Opens interactive IPython shell """
    click.echo('you can use current ctx, requests, provider')
    user_ns = {
        'provider': ctx.provider,
        'ctx': ctx,
        'requests': ctx.provider.requests,
    }

    try:
        import pandas as pd
        click.echo('import pandas as pd')
        user_ns['pd'] = pd
    except ImportError:
        pass

    try:
        from IPython import start_ipython
        start_ipython(
            argv=[],
            user_ns=user_ns
        )
    except ImportError:
        click.echo('You need to install rebotics_sdk[shell] and you also required to use Python3.5+')


@click.command()
@click.option('-l', '--locate', is_flag=True, default=False)
@pass_rebotics_context
def roles(ctx, locate):
    """List available roles to use"""
    if locate:
        click.echo(ctx.config_path)
        return
    config = ctx.config
    if ctx.format == 'json':
        click.echo(pformat(config.config, indent=2))
    elif ctx.format == 'id':
        for key, value in config.config.items():
            click.echo(key)
    elif ctx.format == 'table':
        role_configurations = []
        for key, value in config.items():
            data = value
            data['role'] = key
            role_configurations.append(data)
        click.echo(format_full_table(role_configurations, 100))


def option_default_from_configurations(conf_name):
    class _Option(click.Option):
        def get_default(self, ctx):
            # try to fetch configuration from conf_name
            if ctx.parent is None:
                raise click.ClickException('This command is supposed to be used as sub command.')
            role = ctx.parent.params.get('role')
            role_config = ctx.obj.config.get(role)
            if role_config is not None:
                self.default = role_config.get(conf_name)
            return super(_Option, self).get_default(ctx)

    return _Option


@click.command()
@click.option('-h', '--host', help='App host', prompt=True, cls=option_default_from_configurations('host'))
@click.option('-u', '--user', prompt=True, cls=option_default_from_configurations('username'))
@click.option('--password', prompt=True, hide_input=True, confirmation_prompt=False)
@pass_rebotics_context
def configure(ctx, user, password, host):
    """Fetch token, save it and use the role of it."""
    provider = ctx.provider_class(host=host)
    ctx.update_configuration(
        host=host,
    )
    try:
        response = provider.token_auth(user, password)
    except ProviderHTTPClientException as e:
        raw_message = e.response.content.decode('utf-8')
        if '2FA' in raw_message or 'code' in raw_message:
            click.echo('Please enter the verification code from email.')
            token = input('Verification code from email: ')
            response = provider.token_auth(user, password, verification_code=token)
        else:
            click.echo("Failed to login because of %s" % str(e))
            return

    ctx.update_configuration(
        host=host,
        date=datetime.now().strftime('%c'),
        **response.json()
    )
    click.echo('Saved configuration for {} in {}'.format(ctx.role, ctx.config_provider.filepath), err=True)
    ctx.format_result(response.json())


@click.command()
@click.option('-h', '--host', help='App host', prompt=True, cls=option_default_from_configurations('host'))
@click.option('--token', prompt=True, hide_input=True, confirmation_prompt=False)
@pass_rebotics_context
def set_token(ctx, host, token):
    """
    Set raw authorization token
    """
    ctx.update_configuration(
        host=host,
        date=datetime.now().strftime('%c'),
        token=token,
    )
    click.echo('Saved configuration for {} in {}'.format(ctx.role, ctx.config_provider.filepath), err=True)


@click.command()
@click.argument('source_file', type=click.File('r'))
@pass_rebotics_context
def run(ctx, source_file):
    click.echo(six.exec_(source_file.read(), {
        'ctx': ctx,
        'provider': ctx.provider,
    }))


@main.command()
@click.argument('codeword')
@click.argument('param')
def signature(codeword, param):
    click.echo(hashlib.md5(str(codeword + param).encode('utf-8')).hexdigest())


@main.command()
@click.argument('compose_file', type=click.Path(exists=True, dir_okay=False))
def clear_ports(compose_file):
    try:
        import psutil
    except ImportError:
        raise ImportError('Try to install psutil')

    compose_absolute_path = pathlib.Path(compose_file).absolute()
    with s.Popen("docker-compose -f {} config".format(compose_absolute_path), stdout=s.PIPE) as proc:
        config = proc.stdout.read().decode('utf-8')
    conf = yaml.load(config, Loader=yaml.SafeLoader)

    for service_name, value in conf['services'].items():
        for ports_map in value.get('ports', []):
            try:
                published = ports_map['published']
            except Exception as exc:
                continue

            connections = [p for p in psutil.net_connections() if p.laddr.port == published]

            for con in connections:
                click.echo("{}:{} {}".format(service_name, ports_map, con))
                try:
                    proc = psutil.Process(con.pid)
                    proc.kill()
                except Exception as exc:
                    pass



from django.dispatch import receiver

from rebotics_sdk.hook.signals import (
    IMPORT_FINISH,
    PROCESSING_FINISH,
    IMPORT_FINISH_ERROR,
    IMPORT_FINISH_SUCCESS,
    PROCESSING_FINISH_ERROR,
    PROCESSING_FINISH_SUCCESS,
)


@receiver(IMPORT_FINISH)
def handle_message(sender, retailer, event, payload, **kwargs):
    print('sender: {}'.format(sender))
    print('{} {} {}'.format(retailer, event, payload))


@receiver(PROCESSING_FINISH)
def handle_message(sender, retailer, event, payload, **kwargs):
    print('sender: {}'.format(sender))
    print('{} {} {}'.format(retailer, event, payload))


@receiver(IMPORT_FINISH_ERROR)
def handle_message(sender, retailer, event, payload, **kwargs):
    print('sender: {}'.format(sender))
    print('{} {} {}'.format(retailer, event, payload))


@receiver(IMPORT_FINISH_SUCCESS)
def handle_message(sender, retailer, event, payload, **kwargs):
    print('sender: {}'.format(sender))
    print('{} {} {}'.format(retailer, event, payload))


@receiver(PROCESSING_FINISH_ERROR)
def handle_message(sender, retailer, event, payload, **kwargs):
    print('sender: {}'.format(sender))
    print('{} {} {}'.format(retailer, event, payload))


@receiver(PROCESSING_FINISH_SUCCESS)
def handle_message(sender, retailer, event, payload, **kwargs):
    print('sender: {}'.format(sender))
    print('{} {} {}'.format(retailer, event, payload))




class RCDB:
    EXPORT_CREATED = 'EXPORT_CREATED'
    EXPORT_DONE = 'EXPORT_DONE'
    EXPORT_IN_PROGRESS = 'EXPORT_IN_PROGRESS'
    EXPORT_ERROR = 'EXPORT_ERROR'

    EXPORT_STATUS_CHOICES = (
        (EXPORT_CREATED, "Export created"),
        (EXPORT_IN_PROGRESS, "Export in progress"),
        (EXPORT_DONE, "Export Done"),
        (EXPORT_ERROR, "Export error"),
    )

    ZIP = 'zip'
    CLASSIFICATION_DATABASE = 'rcdb'
    VIRTUAL_CLASSIFICATION_DATABASE = 'vrcdb'
    JSON = 'json'

    EXTENSION_CHOICES = (
        (ZIP, 'zip archive'),
        (CLASSIFICATION_DATABASE, 'Rebotics database format'),
        (VIRTUAL_CLASSIFICATION_DATABASE, 'Rebotics virtual database format'),
        (JSON, "Flat json of image urls and labels"),
    )

    FEATURE_VECTORS = 'feature_vectors'
    FEATURE_PREVIEWS = 'feature_previews'
    HEAT_MAP = 'heat_map'
    SIZES = 'sizes'
    BACKUP_TYPES = (
        (FEATURE_VECTORS, 'Feature Vectors'),
        (FEATURE_PREVIEWS, 'Feature vectors + Previews'),
        (HEAT_MAP, 'Heat map information'),
        (SIZES, 'Sizes information'),
    )


class DownloadImagesFormat(object):
    FULL = "full"
    IMAGES = "images"
    ANNOTATIONS = "annotations"
    OPTIONS = [
        FULL,
        IMAGES,
        ANNOTATIONS,
    ]



import os

import click

from rebotics_sdk.cli.common import configure, shell, roles, set_token
from rebotics_sdk.cli.utils import read_saved_role, process_role, ReboticsCLIContext, app_dir, pass_rebotics_context
from rebotics_sdk.providers import ProviderHTTPServiceException
from rebotics_sdk.providers.hawkeye import HawkeyeProvider


@click.group()
@click.option('-f', '--format', default='table', type=click.Choice(['table', 'id', 'json']), help='Result rendering')
@click.option('-v', '--verbose', is_flag=True, help='Enables verbose mode')
@click.option('-c', '--config', type=click.Path(), default='hawkeye.json', help="Specify what config.json to use")
@click.option('-r', '--role', default=lambda: read_saved_role('hawkeye'), help="Key to specify what hawkeye to use")
@click.version_option()
@click.pass_context
def api(ctx, format, verbose, config, role):
    """
    Admin CLI tool to communicate with Hawkeye API
    """
    process_role(ctx, role, 'hawkeye')
    ctx.obj = ReboticsCLIContext(
        role,
        format,
        verbose,
        os.path.join(app_dir, config),
        provider_class=HawkeyeProvider
    )


@api.command(name='heartbeat')
@click.option('-c', '--camera', required=True, help='Shelf Camera ID to save its heartbeat data')
@click.option('-b', '--battery', required=True, help='Battery status of the camera', type=click.FLOAT)
@click.option('-f', '--free-storage', 'free_storage', required=True, help='Free storage of the Camera',
              type=click.FLOAT)
@click.option('-u', '--used-storage', 'used_storage', required=True, help='Used storage of the Camera',
              type=click.FLOAT)
@pass_rebotics_context
def create_camera_heartbeat(ctx, camera, battery, free_storage, used_storage):
    """Saves camera heartbeat data and returns the camera's list of actions"""
    try:
        if ctx.verbose:
            click.echo('Calling create camera heartbeat')
        result = ctx.provider.save_camera_heartbeat(
            camera,
            battery,
            free_storage,
            used_storage
        )
        click.echo(result)
    except ProviderHTTPServiceException as exc:
        raise click.ClickException(str(exc))


@api.command(name='fixture')
@click.option('-i', '--store', required=True, help='Store ID of a fixture', type=click.STRING)
@click.option('-a', '--aisle', required=True, help='Aisle of a fixture', type=click.STRING)
@click.option('-s', '--section', required=True, help='Section of a fixture', type=click.STRING)
@click.option('-r', '--retailer', required=True, help='Retailer of a fixture', type=click.STRING)
@pass_rebotics_context
def create_fixture(ctx, store, aisle, section, retailer):
    """Saves Fixture object and returns ..."""
    try:
        if ctx.verbose:
            click.echo('Calling create fixture')
        result = ctx.provider.save_fixture(
            store,
            aisle,
            section,
            retailer
        )
        click.echo(result)
    except ProviderHTTPServiceException as exc:
        raise click.ClickException(str(exc))


@api.command(name='fixtures')
@pass_rebotics_context
def get_list_fixtures(ctx):
    """Shows list of fixtures"""
    try:
        if ctx.verbose:
            click.echo('Calling get fixtures')
        result = ctx.provider.get_fixtures()
        click.echo(result)
    except ProviderHTTPServiceException as exc:
        raise click.ClickException(str(exc))


@api.command(name='fixture-delete')
@click.option('-i', '--pk', required=True, help='ID of a fixture to be deleted', type=click.STRING)
@pass_rebotics_context
def delete_fixture(ctx, pk):
    """Shows list of fixtures"""
    try:
        if ctx.verbose:
            click.echo('Calling delete fixture')
        result = ctx.provider.delete_fixture(
            pk=pk
        )
        click.echo(result)
    except ProviderHTTPServiceException as exc:
        raise click.ClickException(str(exc))


@api.command(name='retail')
@click.option('-c', '--codename', help='Codename of the retailer', type=click.STRING)
@click.option('-t', '--token', help='Token of the retailer', type=click.STRING)
@pass_rebotics_context
def create_retailer(ctx, codename, token):
    """Create Retailer object"""
    try:
        if ctx.verbose:
            click.echo('Calling create retailer')
        result = ctx.provider.save_retailer(
            codename,
            token
        )
        ctx.format_result(result)
        click.echo(result)
    except ProviderHTTPServiceException as exc:
        raise click.ClickException(str(exc))


@api.command(name='camera')
@click.option('-c', '--camera', required=True, help='Camera ID of the Shelf Camera', type=click.STRING)
@click.option('-a', '--added', required=True, help='Who added the camera (ID)', type=click.INT)
@click.option('-f', '--fixture', required=True, help='Fixture ID of the camera', type=click.INT)
@pass_rebotics_context
def create_shelf_camera(ctx, camera, added, fixture):
    """Saves ShelfCamera object"""
    try:
        if ctx.verbose:
            click.echo('Calling create shelf camera')
        result = ctx.provider.create_shelf_camera(
            camera,
            added,
            fixture
        )
        ctx.format_result(result)
        click.echo(result)
    except ProviderHTTPServiceException as exc:
        raise click.ClickException(str(exc))


@api.command(name='cameras')
@pass_rebotics_context
def get_list_shelf_cameras(ctx):
    """Shows list of shelf cameras"""
    try:
        if ctx.verbose:
            click.echo('Calling create shelf camera')
        result = ctx.provider.get_shelf_cameras()
        ctx.format_result(result)
        click.echo(result)
    except ProviderHTTPServiceException as exc:
        raise click.ClickException(str(exc))


@api.command(name='capture')
@click.option('-c', '--camera', required=True, help='camera_id of the Capture object', type=click.STRING)
@click.option('-f', '--filekey', required=True, help='File key to be set in Capture object', type=click.STRING)
@click.option('-b', '--bucket-name', required=True, help='Bucket name to bet set in Capture object', type=click.STRING)
@pass_rebotics_context
def create_capture(ctx, camera, filekey, bucket_name):
    """Create Capture object with file_id to be configured, and ignored MEDIA_ROOT configuration"""
    click.echo('enters')
    try:
        if ctx.verbose:
            click.echo('Calling create capture')
        result = ctx.provider.save_capture(
            camera,
            filekey,
            bucket_name
        )
        ctx.format_result(result)
        click.echo(result)
    except ProviderHTTPServiceException as exc:
        raise click.ClickException(str(exc))


api.add_command(shell, 'shell')
api.add_command(roles, 'roles')
api.add_command(configure, 'configure')
api.add_command(set_token, 'set_token')



import logging

from .base import ReboticsBaseProvider, remote_service

logger = logging.getLogger(__name__)


class DatasetProvider(ReboticsBaseProvider):
    @remote_service('/api/v1/')
    def api_root(self):
        return self.session.get()

    @remote_service('/api/v1/token-auth/', raw=True)
    def token_auth(self, username, password, verification_code=None):
        payload = dict(
            username=username,
            password=password,
        )
        if verification_code is not None:
            payload['verification_code'] = verification_code

        response = self.session.post(data=payload)
        self.headers['Authorization'] = 'Token %s' % response.json()['token']
        return response

    @remote_service('/api/v1/import_training_data/', timeout=10000)
    def import_training_data(self, data):
        return self.session.post(json=data)

    @remote_service('/api/v1/detection-images/{id}/')
    def get_detection_image_by_id(self, image_id):
        return self.session.get(id=image_id)



import json

import six

from rebotics_sdk.providers.utils import hash_file, accept_file
from .base import ReboticsBaseProvider, remote_service, PageResult, UniqueRequestProxy
from ..advanced.flows import FeatureVectorFlow


class RetailerProvider(ReboticsBaseProvider):
    proxy_class = UniqueRequestProxy
    proxy_config = {
        'request_uid_header_name': 'REB-CLIENT-REQUEST-UID'
    }

    @remote_service('/api/v4/')
    def api_v4_root(self, **kwargs):
        return self.session.get()

    @remote_service('/api/v4/version/')
    def version(self, *kwargs):
        return self.session.get()

    @remote_service('/api/v4/token-auth/', raw=True)
    def token_auth(self, username, password, **kwargs):
        payload = dict(
            username=username,
            password=password,
        )
        if 'verification_code' in kwargs.keys():
            payload['verification_code'] = kwargs['verification_code']
        response = self.session.post(json=payload)
        self.set_token(response.json()['token'])
        return response

    def api_login(self, username, password):
        return self.token_auth(username, password)

    def get_token(self, username, password):
        response = self.api_login(username, password)
        return response['token']

    @remote_service('/api/v4/processing/actions/{id}/requeue/')
    def requeue(self, processing_action_id, requeue_type=None, **kwargs):
        data = {}

        if requeue_type is not None:
            data['requeue_type'] = requeue_type

        return self.session.post(id=processing_action_id, data=data)

    @remote_service('/api/v4/processing/actions/{id}/cancel/')
    def cancel(self, processing_action_id, **kwargs):
        return self.session.post(id=processing_action_id)

    @remote_service('/api/v4/processing/actions/{id}/recalculate/')
    def recalculate(self, processing_action_id, **kwargs):
        return self.session.post(id=processing_action_id)

    @remote_service('/api/v4/processing/actions/{id}/')
    def processing_action_detail(self, processing_action_id, **kwargs):
        return self.session.get(id=processing_action_id)

    @remote_service('/api/v4/processing/actions/{id}/realogram/')
    def processing_action_realogram_detail(self, processing_action_id, **kwargs):
        return self.session.get(id=processing_action_id)

    @remote_service('/api/v4/processing/actions/{id}/')
    def processing_action_delete(self, processing_action_id, **kwargs):
        return self.session.delete(id=processing_action_id)

    @remote_service('/api/v4/processing/actions/{id}/send_oos_notification/')
    def send_oos_notification(self, processing_action_id, force=False, **kwargs):
        return self.session.post(id=processing_action_id, data={
            'force': force
        })

    @remote_service(
        '/api/v4/processing/actions/',
        request_uid_value=lambda **kwargs: kwargs['json'].pop('_request_uid', None))
    def create_processing_action(
        self, store_id, files, input_type='image',
        store_planogram=None,
        aisle=None, section=None,
        klt_json=None, lens_used=None,
        **kwargs
    ):
        """
        Create processing action for store using list of files
        :param int store_id:
        :param list<int> files:
        :param str input_type:
        :param int store_planogram:
        :param str aisle:
        :param str section:
        :param klt_json:
        :param lens_used:
        :param kwargs: all other fields that are sent to reference
        :return:
        """
        for f_ in files:
            assert isinstance(f_, int), "Should send IDs of uploaded files"

        assert input_type in ('video', 'image', 'keyframe'), \
            'Need to get video, image or keyframe, but %s supplied' % input_type

        data = {
            'store': store_id,
            'files': files,
            'input_type': input_type
        }
        if store_planogram:
            assert isinstance(store_planogram, int)
            data['store_planogram'] = store_planogram
        if aisle:
            assert isinstance(aisle, six.string_types)
            data['aisle'] = aisle
        if section:
            assert isinstance(section, six.string_types)
            data['section'] = section

        if klt_json:
            # encode KLT to string as it is accepted in API
            if isinstance(klt_json, dict):
                klt = json.dumps(klt_json)
            elif isinstance(klt_json, six.string_types):
                klt = klt_json
            else:
                # TODO: add path support and IO support
                raise ValueError('Unsupported value type for KLT.jsons')
            data['klt_json'] = klt

        if lens_used is not None:
            data['lens_used'] = lens_used

        for k, v in kwargs.items():
            data[k] = v

        return self.session.post(json=data)

    @remote_service('/api/v4/processing/upload/')
    def processing_upload(self, store_id, input_file, input_type='image', **kwargs):
        """Upload files that can be used for processing action creation"""
        assert isinstance(store_id, int), 'You need to specify store_id as int'
        assert input_type in (
            'image', 'video', 'keyframe'
        ), "File type should one of (image|video|keyframe) not %s" % input_type

        checksum = hash_file(input_file)
        input_file.seek(0)

        return self.session.post(data={
            'checksum': checksum,
            'input_type': input_type,
            'store': store_id
        }, files={
            'file': input_file
        })

    @remote_service('/api/v4/processing/upload/by-reference/')
    def processing_upload_by_reference(self, file_url, input_type='image', **kwargs):
        assert input_type in (
            'image', 'video', 'keyframe'
        ), "File type should one of (image|video|keyframe) not %s" % input_type

        return self.session.post(
            data={
                'input_type': input_type,
                'ref_url': file_url
            }
        )

    @remote_service('/api/v4/processing/upload/request/')
    def processing_upload_request(self, filename, input_type='image', **kwargs):
        return self.session.post(
            json={
                'filename': filename,
                'input_type': input_type
            }
        )

    @remote_service('/api/v4/processing/upload/request/{id}/finish/')
    def notify_processing_upload_finished(self, id_, **kwargs):
        return self.session.post(id=id_)

    @remote_service('/api/v4/export/products/')
    def export_products(self, products_filter=None, page=1, **kwargs):
        params = {
            'page': page,
            'page_size': 100,
        }
        if products_filter:
            params['filter'] = products_filter
        return PageResult(self.session.get(params=params))

    @remote_service('/api/v4/export/products/{product_id}/previews/')
    def get_product_previews(self, product_id, is_removed=None, page=1, **kwargs):
        params = {
            'page': page,
        }
        if is_removed is not None:
            assert isinstance(is_removed, bool)
            params['is_removed'] = is_removed

        return PageResult(self.session.get(
            product_id=product_id,
            params=params
        ))

    @remote_service('/api/v4/export/products/previews/', timeout=30000)
    def get_product_previews_bulk(self, product_id_list, **kwargs):
        filter_params = {'id': product_id_list}
        return self.session.post(data=filter_params)

    @remote_service('/api/v4/products/previews/')
    def upload_product_preview(self, upc, image_file, **kwargs):
        return self.session.post(data={
            'product_unique_id': upc,
        }, files={
            'image': image_file,
        })

    @remote_service('/api/v4/products/previews/delete/')
    def delete_product_previews(self, upc, **kwargs):
        return self.session.post(data={
            'product_unique_id': upc
        })

    @remote_service('/api/v4/products/search/')
    def get_product_info_by_upc(self, upc_list, **kwargs):
        if not all(upc.isdigit() for upc in upc_list):
            raise ValueError("Supplied non numeric code in list")

        return PageResult(self.session.post(json={
            'plu': upc_list
        }))

    @remote_service('/api/v4/product/previews/{preview_id}/add_feature_reference/')
    def set_product_preview_feature_id(self, preview_id, feature_id, facenet, **kwargs):
        return self.session.post(
            preview_id=preview_id,
            data={
                'feature_id': feature_id,
                'facenet': facenet,
            }
        )

    @remote_service('/api/v4/processing/actions/')
    def processing_action_list(self, store=None, status=None, aisle=None, section=None, page=1, page_size=100,
                               **kwargs):
        """Request processing action list"""

        def set_param(params, key, value):
            if value is not None:
                params[key] = value

        params = {
            'page': page,
            'page_size': page_size,
        }

        set_param(params, 'store', store)
        set_param(params, 'status', status)
        set_param(params, 'aisle', aisle)
        set_param(params, 'section', section)
        params.update(kwargs)

        data = self.session.get(params=params)
        return PageResult(data)

    @remote_service('api/v4/imports/planogram/assign/')
    def assign_planogram_with_file(self, assign_file, deactivate_old_planograms=False, **kwargs):

        file_io = None

        result = self.session.post(data={
            'deactivate_old_planograms': deactivate_old_planograms
        }, files={
            assign_file: file_io
        })

        file_io.close()
        return result

    @remote_service('api/v4/imports/planogram/')
    def import_planogram(
        self,
        planogram_file, title, description='',
        mappings=None,
        vertical_direction='top-to-bottom',
        horizontal_direction='left-to-right',
        horizontal_direction_shelf_bay='left-to-right',
        is_zig_zag=False,
        convert_upc=False,
        **kwargs
    ):
        """
        This will also close your file after transmission

        :param planogram_file: Planogram file as path, as StringIO/ByteIO, as URL
        :param str title: Planogram title, that you can use
        :param str description:
        :param dict mappings:
        :param str vertical_direction:
        :param str horizontal_direction:
        :param str horizontal_direction_shelf_bay:
        :param bool is_zig_zag:
        :param bool convert_upc: This flag will indicate that we need to convert what ever UPC to UPC-A format.
                                Please do not check if you provide EAN or other than UPC code.
        :return:
        """
        schema = self.session.options()
        post_schema = schema['actions']['POST']
        mapping_fields = post_schema['mappings']['children']
        option_fields = post_schema['options']['children']
        default_required_mappings = {key: key for key, value in mapping_fields.items() if value['required']}
        if mappings is None:
            mappings = default_required_mappings

        assert isinstance(mappings, dict)
        for key in default_required_mappings:
            if key not in mappings:
                raise AssertionError('You have not provided required key %s for API' % key)

        def get_choice_values(fields, key):
            return (choice['value'] for choice in fields[key]['choices'])

        assert vertical_direction in get_choice_values(option_fields, 'vertical_direction')
        assert horizontal_direction in get_choice_values(option_fields, 'horizontal_direction')
        assert horizontal_direction_shelf_bay in get_choice_values(option_fields, 'horizontal_direction_shelf_bay')
        assert isinstance(is_zig_zag, bool)
        assert isinstance(convert_upc, bool)
        assert isinstance(title, str), 'You need to specify title, instead %s with type %s was provided' % (
            title, type(title))
        assert isinstance(description, str), 'You need to specify description'

        file_io = accept_file(self, planogram_file)

        file_io.seek(0)

        response = self.session.post(json={
            'mappings': mappings,
            'options': {
                'vertical_direction': vertical_direction,
                'horizontal_direction': horizontal_direction,
                'horizontal_direction_shelf_bay': horizontal_direction_shelf_bay,
                'is_zig_zag': is_zig_zag,
                'convert_upc': convert_upc,
            },
            'title': title,
            'description': description,
        },
            files={
                'planogram_file': file_io
            })
        file_io.close()

        return response

    @remote_service('api/v4/imports/planogram-assign-file/')
    def assign_planogram(self, assign_file, deactivate=False, **kwargs):
        """
        Uploads an assign planogram file to retailer. Be careful to upload it.
        :param assign_file: file-like object
        :param bool deactivate: flag to deactivate old existing store planograms
        :return:
        """
        assert isinstance(deactivate, bool)
        file_io = accept_file(self, assign_file)
        file_io.seek(0)

        response = self.session.post(data={
            'deactivate_old_planogram': deactivate
        }, files={
            'assign_file': file_io
        })

        file_io.close()
        return response

    @remote_service('api/v4/store_aisles/{store_id}')
    def get_store_aisles(self, store_id, **kwargs):
        return self.session.get(store_id=store_id)

    @remote_service('api/v4/stores/')
    def get_stores(self, **kwargs):
        return self.session.get()

    @remote_service('api/v4/processing_aisles/{processing_id}/')
    def get_processing_aisles(self, processing_id, **kwargs):
        return self.session.get(processing_id=processing_id)

    @remote_service('api/v4/users/')
    def user_list(self, **kwargs):
        return self.session.get()

    @remote_service('api/v4/users/{username}/subscriptions/')
    def user_subscriptions(self, username, **kwargs):
        return self.session.get(username=username)

    @remote_service('api/v4/users/{username}/subscriptions/')
    def user_subscriptions_create(self, username, store, aisle, section, **kwargs):
        return self.session.post(
            username=username,
            data={
                'store': store,
                'aisle': aisle,
                'section': section,
            }
        )

    @remote_service('api/v1/master-data/file-upload/')
    def v1_upload_file(self, file, **kwargs):
        return self.session.post(
            data={
                'description': 'preview upload from SDK',
            },
            files={
                'file': file
            }
        )

    @remote_service('api/v1/master-data/products/{product_id}/images/')
    def v1_upload_preview(self, product_id, image_id, **kwargs):
        return self.session.post(
            product_id=product_id,
            json={
                'image': image_id
            }, content_type='application/json')

    @remote_service('api/v4/processing/actions/{id}/spacial')
    def get_spacial_data(self, scan_id, **kwargs):
        """Can yield permission error. Internal usage intended"""
        return self.session.get(id=scan_id)

    @remote_service('api/v4/processing/actions/export-to-dataset/')
    def export_to_dataset(self, scan_ids, batch_size=5, **kwargs):
        return self.session.post(json={
            'scans': sorted(list(set(scan_ids))),
            'batch_size': batch_size,
        })

    @remote_service('/api/v1/rebotics/rpc/export-feature-vector/')
    def feature_vectors_export(self):
        return FeatureVectorFlow(self, self.session)

    @remote_service('/api/v1/realograms/scan-batch/bin/')
    def create_scan_batch_bin(self, items):
        # TODO: move to pydantic
        required_keys = {
            "imageName": str,
            "timestamp": str,
            "sensorName": str,
            "clubId": str,
            "annotationVersion": int,
            "locationLabels": {
                "area": (str, False),
                "aisle": (str, False),
                "section": (str, False),
            },
            "pose": {
                "x": float,
                "y": float,
                "theta": (float, False)
            },
            "locationId": (str, False),
            "href": str
        }

        def _validate_field(obj, key, conf):
            value = obj.get(key)
            if isinstance(value, dict):
                return _validate_field(value, key, conf[key])

            required_ = True
            if isinstance(conf, tuple):
                type_, required_ = conf
            else:
                type_ = conf

            if value is None:
                if required_:
                    raise ValueError("Field {} is required".format(key))
            else:
                if isinstance(value, type_):
                    raise ValueError("Field {} does not match required type".format(key, str(type_)))

        for entry in items:
            if not isinstance(entry, dict):
                raise ValueError("Entry should be dict")
            for expected_key, conf in required_keys.items():
                _validate_field(entry, expected_key, conf)

        return self.session.post(json=items)



# coding=utf-8
import csv
import datetime
import json
import os
import pathlib
from functools import partial
from multiprocessing.pool import Pool
from time import sleep

import click
import humanize
import tqdm
from prettytable import PrettyTable

from rebotics_sdk.advanced import remote_loaders
from rebotics_sdk.cli.common import shell, configure, roles, set_token
from .renderers import format_full_table
from .utils import ReboticsCLIContext, app_dir, pass_rebotics_context, read_saved_role, process_role
from ..advanced.flows import FileUploadError
from ..advanced.packers import ClassificationDatabasePacker, ClassificationDatabaseException, \
    DuplicateFeatureVectorsException
from ..constants import RCDB
from ..providers import AdminProvider, RetailerProvider
from ..utils import uri_validator, download_and_return_id, get_filename_from_url


@click.group()
@click.option('-f', '--format', default='table', type=click.Choice(['table', 'id', 'json']), help='Result rendering')
@click.option('-v', '--verbose', is_flag=True, help='Enables verbose mode')
@click.option('-c', '--config', type=click.Path(), default='admin.json', help="Specify what config.json to use")
@click.option('-r', '--role', default=lambda: read_saved_role('admin'), help="Key to specify what admin to use")
@click.version_option()
@click.pass_context
def api(ctx, format, verbose, config, role):
    """
    Admin CLI tool to communicate with dataset API
    """
    process_role(ctx, role, 'admin')
    ctx.obj = ReboticsCLIContext(
        role,
        format,
        verbose,
        os.path.join(app_dir, config),
        provider_class=AdminProvider
    )


def get_retailer_version_task(retailer_dict):
    retailer_provider = RetailerProvider(host=retailer_dict['host'], retries=1, timeout=5)
    try:
        response = retailer_provider.version()
        version = response['version']
        uptime = humanize.naturaldelta(datetime.timedelta(seconds=int(response['uptime'])))
    except Exception:
        version = 'not working'
        uptime = '---'

    d = [
        retailer_dict['codename'],
        retailer_dict['title'],
        version,
        uptime,
        retailer_dict['host'],
    ]
    return d


@api.command()
@click.argument('retailer')
@pass_rebotics_context
def configurations(ctx, retailer):
    """
        Fetch retailer configurations variables
    """
    try:
        res = ctx.provider.get_configurations(retailer)
        ctx.format_result(res)
    except Exception as exc:
        raise click.ClickException(exc)


@api.command()
@click.option('-n', '--notify', is_flag=True)
@click.option('-d', '--delay', type=click.INT, default=60)
@pass_rebotics_context
def retailer_versions(ctx, notify, delay):
    """Fetch retailer versions and their meta information"""
    click.echo("For better experience use: \n"
               "https://versions.fyn.rocks/ and https://tracker.fyn.rocks/")

    if notify:
        if ctx.verbose:
            click.echo('Using notify option', err=True)

        try:
            from pynotifier import Notification
            Notification(
                title='Subscribed to the notifications',
                description='You will receive notifications for retailer updates',
            ).send()
        except ImportError:
            raise click.ClickException("You can't use notify function")

    provider = ctx.provider
    if ctx.verbose:
        click.echo('Fetching info from rebotics admin.', err=True)
    retailers = provider.get_retailer_list()
    prev_results = []
    results = []
    pool = Pool(len(retailers))

    while True:
        try:
            if ctx.verbose:
                click.echo('Fetching the retailer versions', err=True)
            results = pool.map(get_retailer_version_task, retailers)

            if not notify:
                break

            for prev_result in prev_results:
                retailer_codename = prev_result[0]
                previous_version = prev_result[2]
                for result in results:
                    if result[0] == retailer_codename:
                        current_version = result[2]
                        if previous_version != current_version:
                            notification_message = 'Retailer {} updated from version {} to {}'.format(
                                retailer_codename,
                                previous_version,
                                current_version
                            )
                            click.echo(notification_message)
                            Notification(
                                title=notification_message,
                                description='Current uptime is: {}'.format(result[3]),
                                duration=30,
                                urgency=Notification.URGENCY_CRITICAL,
                            ).send()
            del prev_results
            prev_results = results
            sleep(delay)
        except KeyboardInterrupt:
            break

    table = PrettyTable()
    table.field_names = ['codename', 'title', 'version', 'uptime', 'host']
    for result in results:
        table.add_row(result)
    click.echo(table)


@api.command()
@click.argument('retailer')
@pass_rebotics_context
def models(ctx, retailer):
    """
        Fetch and display retailer NN configurations
    """
    try:
        ctx.format_result(
            ctx.provider.get_retailer_tf_models(codename=retailer)
        )
    except Exception as exc:
        raise click.ClickException(exc)


@api.command()
@click.argument('retailer', type=click.STRING)
@click.argument('url', type=click.STRING)
@pass_rebotics_context
def set_retailer_url(ctx, retailer, url):
    try:
        ctx.provider.update_host(retailer, url)
    except Exception as exc:
        raise click.ClickException(str(exc))
    else:
        click.echo('Set new host for retailer %s' % retailer)


@api.command()
@click.option('-r', '--retailer', help="Retailer codename", prompt=True)
@click.option('-m', '--model', help="Model codename", prompt=True)
@click.option('-i', '--images', type=click.Path(exists=True, file_okay=False, dir_okay=True))
@click.option('-l', '--labels', type=click.Path(exists=True, file_okay=True, dir_okay=False))
@click.option('-f', '--features', type=click.Path(exists=True, file_okay=True, dir_okay=False))
@click.option('-c', '--check-duplicates', is_flag=True, help="MD5 checker simple")
@pass_rebotics_context
def pack_and_import_classification_db(ctx, retailer, model, images, labels, features, check_duplicates):
    """Pack classification database to single .rcdb file and import it to the Rebotics Admin

\b
db/
├── custom_folder
│   ├── image_2.png
│   └── image_1.png
├── features.txt
└── labels.txt

    It is a single step command with equivalent of running two commands:

admin pack-classification-db --features features.txt --labels labels.txt --images ./custom_folder/ --target classification.rcdb

admin import_classification_db --retailer delta --model test_code classification.rcdb
    """
    ext = ClassificationDatabasePacker.extension
    with ctx.provider.rcdb.import_flow(retailer, model, ext) as flow:
        if ctx.verbose:
            click.echo("Creating import request: \n"
                       "retailer: {}\n"
                       "model: {}\n"
                       "extension: {}".format(retailer, model, ClassificationDatabasePacker))

        packer = ClassificationDatabasePacker(destination=None, progress_bar=True, check_duplicates=check_duplicates)
        if ctx.verbose:
            click.echo("Packing from provided values: \n"
                       "labels: {labels} \n"
                       "features: {features} \n"
                       "images: {images}".format(labels=labels, features=features, images=images))

        try:
            packed = packer.pack(labels, features, images)
            packed.seek(0)
        except DuplicateFeatureVectorsException as exc:
            handle_duplicates(ctx, exc.duplicates)
            raise click.ClickException(str(exc))
        except ClassificationDatabaseException as exc:
            raise click.ClickException(str(exc))

        if ctx.verbose:
            click.echo("Import request: {}".format(flow.import_request))

        try:
            flow.upload_file(packed, progress_bar=True)
        except FileUploadError as exc:
            click.echo(str(exc))
            raise

        click.echo("Import completed. Reference database id: {}".format(flow.import_request['id']))


@api.command()
@click.option('-r', '--retailer', help="Retailer codename", prompt=True)
@click.option('-m', '--model', help="Model codename", prompt=True)
@click.argument('filepath', type=click.Path(exists=True, file_okay=True, dir_okay=False))
@pass_rebotics_context
def import_classification_db(ctx, retailer, model, filepath):
    """Import rcdb file to Rebotics Admin. Example usage:

    admin import_classification_db --retailer delta --model test_code classification.rcdb

    """
    extension = os.path.split(filepath)[-1].split('.')[-1]
    if ctx.verbose:
        click.echo("Creating import request: \n"
                   "retailer: {}\n"
                   "model: {}\n"
                   "extension: {}".format(retailer, model, extension))

    with ctx.provider.rcdb.import_flow(retailer, model, extension) as flow:
        if ctx.verbose:
            click.echo("Import request: {}".format(flow.import_request))
        with open(filepath, 'rb') as packed:
            flow.upload_file(packed, progress_bar=True)

    click.echo("Import completed. Reference database id: {}".format(flow.import_request['id']))


def handle_duplicates(ctx, duplicates):
    unpacked = [
        [
            {
                'id': entry.index,
                'label': entry.label,
                'filename': entry.filename
            }
            for entry in group
        ] for group in duplicates
    ]
    if ctx.format == 'json':
        click.echo(json.dumps(unpacked, indent=2))
    elif ctx.format == 'table':
        for group in unpacked:
            format_full_table(group)
    elif ctx.format == 'id':
        click.echo("Duplicate id. New line separated for group; space separated inside of the group:")
        for group in duplicates:
            click.echo(" ".join([str(entry.index) for entry in group]))
    else:
        for group in duplicates:
            for entry in group:
                click.echo("{entry.index} {entry.label} {entry.filename}".format(entry=entry))
            click.echo('=' * 100)


@api.command()
@click.option('-i', '--images', type=click.Path(exists=True, file_okay=False, dir_okay=True))
@click.option('-l', '--labels', type=click.Path(exists=True, file_okay=True, dir_okay=False))
@click.option('-f', '--features', type=click.Path(exists=True, file_okay=True, dir_okay=False))
@click.option('-c', '--check-duplicates', is_flag=True, help="MD5 checker simple")
@click.argument('target', type=click.Path(exists=False, file_okay=True, dir_okay=True), default='.')
@pass_rebotics_context
def pack_classification_db(ctx, images, labels, features, target, check_duplicates):
    """Pack classification database to single .rcdb file

\b
db/
├── custom_folder
│   ├── image_2.png
│   └── image_1.png
├── features.txt
└── labels.txt

    Example usage:
        admin pack-classification-db --features features.txt --labels labels.txt --images ./custom_folder/ --target classification.rcdb

    """
    target = pathlib.Path(target)
    if target.is_dir():
        target = target / "classification_db{}.{}".format(
            datetime.datetime.now().strftime('%Y-%m-%dZ%H%M'),
            ClassificationDatabasePacker.extension
        )
    packer = ClassificationDatabasePacker(
        destination=target,
        progress_bar=True,
        check_duplicates=check_duplicates
    )
    if ctx.verbose:
        click.echo("Packing from provided values: \n"
                   "labels: {labels} \n"
                   "features: {features} \n"
                   "images: {images}".format(labels=labels, features=features, images=images))
    try:
        packed = packer.pack(labels, features, images)
    except DuplicateFeatureVectorsException as exc:
        handle_duplicates(ctx, exc.duplicates)
        raise click.ClickException(str(exc))
    except ClassificationDatabaseException as exc:
        raise click.ClickException(str(exc))
    click.echo('Written to {}'.format(packed))


@api.group(name='rcdb')
def rcdb_group():
    click.echo("RCDB in rebotics admin is deprecated and will be removed in version 0.13.0")


@rcdb_group.command(name="previews-fs")
@click.argument("argument", type=click.STRING)
@click.option('-o', "--output", type=click.Path(exists=True, dir_okay=True, file_okay=False), default=".")
@click.option('-c', '--concurrency', default=8, type=click.INT)
@pass_rebotics_context
def rcdb_previews_fs(ctx, argument, output, concurrency):
    """
    Prepare previews file structure by the id, filepath or file url of the JSON formatted backup
    Make sure that extension is a valid JSON and extension in the admin is set like that
    """
    output = pathlib.Path(output).absolute()
    output.mkdir(parents=True, exist_ok=True)

    filepath = output / "definition.json"

    if argument.isdigit():
        backup_id = int(argument)
        rcdb_data = ctx.provider.rcdb.get(backup_id)
        if rcdb_data['extension'] != RCDB.JSON:
            click.echo("extension is not supported")  # but we will try anyway
            # raise click.ClickException("Extension is not supported")

        # download to a local folder
        remote_loaders.download(rcdb_data['backup'], filepath)

    elif uri_validator(argument):
        # argument is a url
        remote_loaders.download(argument, filepath)

    else:
        filepath = argument

    try:
        with open(filepath, 'rb') as fp:
            definition = json.load(fp)
    except:
        raise click.ClickException(f"File is not supported. Please check it at {filepath}")

    # apply a script
    try:
        import pandas as pd
        df = pd.DataFrame(definition)
        df.to_csv(output / "definition.csv", quoting=csv.QUOTE_NONNUMERIC)
        del df
    except:
        pass

    with Pool(concurrency) as pool:
        for _ in tqdm.tqdm(
            pool.imap_unordered(partial(download_and_return_id, output=output), definition),
            desc=f"Downloading images to {str(output)}",
            total=len(definition),
        ):
            pass
        pool.close()
        pool.join()


@api.group()
def stitching():
    pass


@stitching.command(name='setup')
@click.argument("argument", type=click.STRING)
@click.option('-o', "--output", type=click.Path(dir_okay=True, file_okay=False))
@click.option('-c', '--concurrency', default=8, type=click.INT)
@pass_rebotics_context
def stitching_setup(ctx, argument, concurrency, output):
    """
    argument - id of stitching debug data from admin, or json file
    """
    if output:
        output = pathlib.Path(output).absolute()
    else:
        output = pathlib.Path('.').absolute()
    output = output / f"stitching_{argument.replace(' ', '_')}"
    output.mkdir(parents=True, exist_ok=True)
    ctx.verbose_log(f"Target directory is: {output}")
    filepath = output / "definition.json"

    if argument.isdigit():
        backup_id = int(argument)
        callback_data = ctx.provider.get_core_callback_data(backup_id)
        definition = callback_data['data']

        ctx.verbose_log("Saving definition locally")
        with open(filepath, 'w') as fp:
            json.dump(definition, fp)

    elif uri_validator(argument):
        # argument is a url
        raise click.ClickException("not supported")
    else:
        ctx.verbose_log("Assuming we have a filepath")
        filepath = argument
        try:
            with open(filepath, 'rb') as fp:
                definition = json.load(fp)
        except:
            raise click.ClickException(f"File is not supported. Please check it at {filepath}")

    if 'data' in definition:
        definition = definition['data']

    if ctx.verbose:
        ctx.format_result(definition['context'], force_format='json')

    files_to_load = []  # pair of remote_url to local path

    for call_count, call_data in enumerate(definition['calls']):
        call_folder = output / f"call_{call_count}"
        call_folder.mkdir(parents=True, exist_ok=True)

        input_data = call_data['input']

        stitching_input = {
            **input_data,
            'frame_paths': [],
            'output_file': str(call_folder),
            'result_image': str(call_folder / "local_stitching.jpeg"),
        }

        del stitching_input['frame_urls']
        del stitching_input['annotated_frames']

        # add stitching result image
        try:
            files_to_load.append([
                call_data['image_url'], call_folder / "server_stitched_image.jpeg"
            ])
        except:
            ctx.verbose_log("No stitching image is available")

        if input_data.get('mask_od_url'):
            mask_od_path = call_folder / "mask_od.json"
            stitching_input['mask_od'] = str(mask_od_path)
            files_to_load.append([
                input_data['mask_od_url'], mask_od_path
            ])

        # register input files
        frames_folder = call_folder / "frames"
        frames_folder.mkdir(parents=True, exist_ok=True)

        for frame_url in input_data['frame_urls']:
            frame_path = frames_folder / get_filename_from_url(frame_url)
            stitching_input['frame_paths'].append(str(frame_path))
            files_to_load.append([
                frame_url, frame_path
            ])

        frames_folder = call_folder / "annotated_frames"
        frames_folder.mkdir(parents=True, exist_ok=True)

        annotated_frames = input_data.get('annotated_frames', [])
        if annotated_frames:
            for frame_url in annotated_frames:
                frame_path = frames_folder / get_filename_from_url(frame_url)
                files_to_load.append([
                    frame_url, frame_path
                ])

        ctx.verbose_log(f'Creating a stitching input for run #{call_count} into {call_folder}')
        with open(str(call_folder / "stitching_input.json"), 'w') as fp:
            json.dump(stitching_input, fp)

    ctx.verbose_log(f"Registered files for downloading: {len(files_to_load)}")

    with Pool(concurrency) as pool:
        for _ in tqdm.tqdm(
            pool.starmap(remote_loaders.download, files_to_load),
            desc=f"Downloading files to {str(output)}",
            total=len(files_to_load),
        ):
            pass
        pool.close()
        pool.join()


api.add_command(shell, 'shell')
api.add_command(roles, 'roles')
api.add_command(configure, 'configure')
api.add_command(set_token, 'set_token')



import django.dispatch

EVENTS = (
    ('report.ready', 'Report ready'),
    ('import.finish', 'Import finish'),
    ('processing.finish', 'Processing finish'),
    ('scan_batch.report.ready', 'Scan batch report ready')
)

# events
WEBHOOK_RECEIVED = django.dispatch.Signal()
IMPORT_FINISH = django.dispatch.Signal()
PROCESSING_FINISH = django.dispatch.Signal()
SCAN_BATCH_REPORT_READY = django.dispatch.Signal()

# statuses
IMPORT_FINISH_ERROR = django.dispatch.Signal()
IMPORT_FINISH_SUCCESS = django.dispatch.Signal()
PROCESSING_FINISH_ERROR = django.dispatch.Signal()
PROCESSING_FINISH_SUCCESS = django.dispatch.Signal()

EVENT_SIGNALS = {
    'report.ready': WEBHOOK_RECEIVED,
    'import.finish': WEBHOOK_RECEIVED,
    'processing.finish': PROCESSING_FINISH,
    'scan_batch.report.ready': SCAN_BATCH_REPORT_READY,
}


class SignalEmitter(object):
    def __init__(self, retailer, event, payload):
        self.retailer = retailer
        self.payload = payload
        self.event = event

    def send_signal(self, signal):
        signal.send_robust(
                sender=self,
                retailer=self.retailer,
                event=self.event,
                payload=self.payload
        )

    def emit_status(self):
        if self.event == 'import.finish':
            if self.payload.get('status', None) == 'done':
                self.send_signal(IMPORT_FINISH_SUCCESS)
            elif self.payload.get('status', None) == 'error':
                self.send_signal(IMPORT_FINISH_ERROR)
        elif self.event == 'processing.finish':
            if self.payload.get('status', None) == 'success':
                self.send_signal(PROCESSING_FINISH_SUCCESS)
            elif self.payload.get('status', None) == 'error':
                self.send_signal(PROCESSING_FINISH_ERROR)

    def emit(self):
        self.send_signal(EVENT_SIGNALS[self.event])
        self.emit_status()




from rest_framework import status
from rest_framework.response import Response
from rest_framework.views import APIView

from rebotics_sdk.hook.serializers import WebhookDataSerializer
from rebotics_sdk.hook import signals


class WebhookHandlerViewSet(APIView):
    permission_classes = []

    def post(self, request, *args, **kwargs):
        retailer_code = kwargs['retailer_code']
        serializer = WebhookDataSerializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        emitter = signals.SignalEmitter(
            retailer=retailer_code,
            event=serializer.validated_data['event'],
            payload=serializer.validated_data['payload'],
        )
        emitter.emit()

        return Response(status=status.HTTP_202_ACCEPTED)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, find_packages

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

test_requirements = ['pytest', ]

setup(
    author="Malik Sulaimanov",
    author_email='malik@retechlabs.com',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
    description="Rebotics SDK for communicating with Rebotic Services, API CLI client.",
    entry_points={
        'console_scripts': [
            'admin=rebotics_sdk.cli.admin:api',
            'dataset=rebotics_sdk.cli.dataset:api',
            'retailer=rebotics_sdk.cli.retailer:api',
            'rebm=rebotics_sdk.cli.retailer:api',
            'rebotics=rebotics_sdk.cli.common:main',
            'fvm=rebotics_sdk.cli.fvm:api',
            'hawkeye=rebotics_sdk.cli.hawkeye:api',
        ],
    },
    long_description=readme + '\n\n' + history,
    include_package_data=True,
    keywords='rebotics_sdk',
    name='rebotics_sdk',
    packages=find_packages(exclude=["tests*"]),
    test_suite='tests',
    url='http://retechlabs.com/rebotics/',
    version='0.15.3',
    zip_safe=False,
    install_requires=[
        'requests>=2.21.0',
        'requests[socks]',
        'requests-toolbelt>=0.9.1',
        'six>=1.12.0',
        'more-itertools',
        'tqdm',
        'chardet',
    ],
    tests_require=test_requirements,
    extras_require={
        'hook': [
            'django',
            'djangorestframework'
        ],
        'shell': [
            'ipython>=7.5.0,<8',
            'pandas',
            'pytz',
            'ptable',
            'python-dateutil',
            'humanize',
            'PySocks>=1.7.1',
            'xlrd>=1.2.0',
            'click>=7.0',
            'pyyaml',
        ]
    }
)



import dataclasses
import functools
import io
import json
import logging
import os
import pathlib
import re
import typing
import uuid
import zipfile
from collections import namedtuple, defaultdict
from datetime import datetime
from hashlib import md5
from io import BytesIO, StringIO
from multiprocessing.pool import ThreadPool

import tqdm

import rebotics_sdk
from rebotics_sdk.advanced import remote_loaders

logger = logging.getLogger(__name__)

ClassificationEntry = namedtuple('ClassificationEntry', ['label', 'feature', 'image', 'filename', 'index'])
ImageEntry = namedtuple('ImageEntry', ['filename', 'filepath', 'order'])


class VirtualClassificationEntry(object):
    def __init__(self, label, feature, image_url, index=None):
        self.label = label
        self.feature = feature
        self.image_url = image_url
        self.index = index

    def to_dict(self):
        return {
            'label': self.label,
            'feature': self.feature,
            'image_url': self.image_url,
            'index': self.index
        }

    @classmethod
    def from_dict(cls, d):
        return cls(
            label=d['label'],
            feature=d['feature'],
            image_url=d['image_url'],
            index=d['index']
        )


NUMERIC_VALUE = re.compile(r'\d+')


class ClassificationDatabaseException(Exception):
    pass


class DuplicateFeatureVectorsException(ClassificationDatabaseException):
    def __init__(self, msg, duplicates):
        super(DuplicateFeatureVectorsException, self).__init__(msg)
        self.duplicates = duplicates


def extract_numeric(filename):
    found = NUMERIC_VALUE.findall(filename)
    if len(found) == 0:
        raise ClassificationDatabaseException('Images name should contain numeric value that represents the order. '
                                              'Instead of {} name was used'.format(filename))
    return int(found[0])


class BaseClassificationDatabasePacker(object):
    version = 0
    extension = 'zip'

    def __init__(self, source=None, destination=None, progress_bar=False, check_duplicates=False,
                 *args, **kwargs):
        if source is not None and destination is not None:
            raise ValueError("You can't sent source and destination at a same time.")

        self.source = source

        self.meta_data = {
            'packed': datetime.now().strftime('%c'),
            'sdk_version': rebotics_sdk.__version__,
            'packer_version': self.version
        }
        self.zip_io = None

        if self.source is not None:
            self.read_meta()

        self.check_duplicates = check_duplicates
        self.images = []

        if hasattr(destination, 'write') or hasattr(destination, 'read'):
            self.destination = destination  # file like object
        elif destination is None:
            self.destination = BytesIO()
        elif isinstance(destination, str) or isinstance(destination, pathlib.PurePath):
            # in python 2 it doesn't accept pathlib PurePath, but only os.path or str
            self.destination = str(destination)
            if not self.destination.endswith(self.extension):
                self.destination = "{}.{}".format(self.destination, self.extension)

        self.progress_bar = progress_bar

    def read_lines(self, lines):
        return [x for x in lines.split('\n') if x]

    def zipfile(self, file, **kwargs):
        # if python version is
        params = dict(
            compression=zipfile.ZIP_DEFLATED,
            allowZip64=True,
        )
        #
        # if (sys.version_info.major, sys.version_info.minor) >= (3, 7):
        #     # using the most aggressive compression
        #     params['compresslevel'] = 9  # default is 6

        params.update(kwargs)

        for compression_option in [zipfile.ZIP_DEFLATED, zipfile.ZIP_STORED]:
            try:
                params['compression'] = compression_option
                return zipfile.ZipFile(file, **params)
            except RuntimeError:
                pass

    def __enter__(self):
        if self.zip_io is not None:
            return self

        # check for file to be not closed

        if self.source is not None:
            self.zip_io = self.zipfile(self.source, mode='r')
        elif self.destination is not None:
            self.zip_io = self.zipfile(self.destination, mode='w')
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.zip_io.close()
        self.zip_io = None

    def pack(self, *args, **kwargs):
        raise NotImplementedError()

    def unpack(self):
        raise NotImplementedError()

    def read_meta(self):
        with self as packer:
            self.meta_data = json.load(packer.zip_io.open('meta.json'))


class ZipDatabasePacker(BaseClassificationDatabasePacker):
    """
    Unified file format for packing classification database into single folder
    File format is the zip based archive with following :

    labels.txt
    features.txt
    meta.json
    """
    version = 0
    extension = 'zip'

    def pack(self, labels, features, *args, **kwargs):
        labels_io = StringIO("\n".join(labels))
        features_io = StringIO("\n".join(features))
        self.meta_data['count'] = len(labels)

        with self.zipfile(self.destination, mode='w') as zip_io:
            zip_io.writestr('labels.txt', labels_io.getvalue())
            zip_io.writestr('features.txt', features_io.getvalue())
            zip_io.writestr('meta.json', json.dumps(self.meta_data))
        return self.destination

    def unpack(self):
        if self.source is None:
            raise ClassificationDatabaseException("Can not unpack with empty source file")

        with self.zipfile(self.source, mode='r') as zip_io:
            try:
                self.meta_data = json.load(zip_io.open('meta.json'))
            except KeyError:
                # meta.json is not presented, working in the compatibility mode
                pass

            labels = self.read_lines(zip_io.read('labels.txt').decode('utf-8'))
            features = self.read_lines(zip_io.read('features.txt').decode('utf-8'))

            for index, (label, feature) in enumerate(zip(labels, features)):
                yield ClassificationEntry(
                    label, feature, None, None, index
                )


class ClassificationDatabasePacker(BaseClassificationDatabasePacker):
    """
    Unified file format for packing classification database into single folder
    File format is the zip based archive with following :

    labels.txt
    features.txt
    meta.json
    images/  - folder with images, this folder can be empty
    """

    version = 1
    extension = 'rcdb'
    images_extensions = {'jpeg', 'png', 'jpg'}

    def extract_meta_data(self, labels_path, features_path, images_folder):
        known_order = set()
        for f in pathlib.Path(images_folder).iterdir():
            if f.is_file():
                extension = f.suffix
                if extension in self.images_extensions:
                    raise ClassificationDatabaseException("Extension is not supported by packer: {}".format(f.name))
                numeric_order = extract_numeric(f.name)
                if numeric_order in known_order:
                    raise ClassificationDatabaseException("Numeric order collision. {}".format(f.name))

                known_order.add(numeric_order)
                self.images.append(ImageEntry(f.name, str(f), numeric_order))

        # repack based on the filename and it's numeric value
        self.images.sort(key=lambda i: i.order)  # sort by numeric value of the filename

        with open(labels_path, 'r') as labels_io, open(features_path, 'r') as features_io:
            labels = self.read_lines(labels_io.read())
            features = self.read_lines(features_io.read())

        labels_count = len(labels)
        features_count = len(features)
        files_count = len(self.images)

        if labels_count != features_count:
            raise ClassificationDatabaseException('Inconsistent count of labels and features. {}/{}'.format(
                labels_count, features_count
            ))

        if labels_count != files_count:
            raise ClassificationDatabaseException(
                'Inconsistent count of labels and features and files. {}/{}/{}'.format(
                    labels_count, features_count, files_count
                ))

        self.meta_data.update({
            'count': features_count,
            'images': [str(img.filename) for img in self.images]
        })

        if not self.check_duplicates:
            return

        # checking for the uniqueness of the FV
        features_map = defaultdict(list)
        for i, fv in tqdm.tqdm(enumerate(features), total=features_count, disable=not self.progress_bar, leave=False):
            md_ = md5(fv.encode('utf-8'))
            features_map[md_.hexdigest()].append(i)
        duplicates = []
        for same_fv_ids in features_map.values():
            if len(same_fv_ids) > 1:
                duplicates.append([
                    ClassificationEntry(
                        labels[id_],
                        features[id_],
                        None,
                        self.images[id_].filename,
                        id_
                    ) for id_ in same_fv_ids
                ])

        if duplicates:
            raise DuplicateFeatureVectorsException(
                "Detected {} duplicate groups.".format(len(duplicates)),
                duplicates=duplicates
            )

    def pack(self, labels, features, images, *args, **kwargs):
        self.extract_meta_data(labels, features, images)

        with self.zipfile(self.destination, mode='w') as zip_io:
            zip_io.write(labels, 'labels.txt')
            zip_io.write(features, 'features.txt')
            zip_io.writestr('meta.json', json.dumps(self.meta_data))

            for img in tqdm.tqdm(
                self.images,
                leave=False,
                disable=not self.progress_bar
            ):
                zip_io.write(img.filepath, str(pathlib.Path('images') / img.filename))

        return self.destination

    def unpack(self):
        """

        :return: generator of label, feature, image_io
        :rtype:self.meta_data
        """
        if self.source is None:
            raise ClassificationDatabaseException("Can not unpack with empty source file")

        with self.zipfile(self.source, mode='r') as zip_io:
            labels = self.read_lines(zip_io.read('labels.txt').decode('utf-8'))
            features = self.read_lines(zip_io.read('features.txt').decode('utf-8'))

            for index, (label, feature, image_name) in enumerate(zip(
                labels,
                features,
                self.meta_data['images']
            )):
                image = zip_io.read('images/{}'.format(image_name))
                yield ClassificationEntry(
                    label, feature, image, image_name, index
                )


@dataclasses.dataclass
class DataEntry:
    label: str
    feature: typing.List[float]
    uuid: uuid.UUID
    image_url: str
    image: typing.Optional[typing.BinaryIO] = None


def _task_unpack_classification_entry(entry: DataEntry, download_image=False):
    if not download_image:
        return entry

    data = dataclasses.asdict(entry)
    data['image'] = buffer = remote_loaders.download(entry.image_url)
    buffer.seek(0)
    return DataEntry(**data)


class VirtualClassificationDatabasePacker(BaseClassificationDatabasePacker):
    version = 2
    extension = 'rcdb'

    def __init__(self, *args, **kwargs):
        self.with_images = kwargs.get('with_images', False)
        self.concurrency = kwargs.get('concurrency', None)
        if self.concurrency is None or self.concurrency < 1:
            self.concurrency = os.cpu_count()
        super(VirtualClassificationDatabasePacker, self).__init__(*args, **kwargs)

    def pack(self, labels, features, uuids, images, *args, **kwargs):
        with self.zipfile(self.destination, mode='w') as zip_io:
            # writing those as a file
            zip_io.write(labels, 'labels.txt')
            zip_io.write(features, 'features.txt')
            zip_io.write(uuids, 'uuids.txt')
            zip_io.write(images, 'images.txt')

            zip_io.writestr('meta.json', json.dumps(self.meta_data))
        return self.destination

    def unpack(self):
        if self.with_images:
            task = functools.partial(_task_unpack_classification_entry, download_image=True)
        else:
            task = _task_unpack_classification_entry

        with ThreadPool(self.concurrency) as pool:
            for batch in self._data_entry_generator():
                yield from pool.map(task, batch)

    def _data_entry_generator(self):
        batch = []  # initial batch
        with self.zipfile(self.source, mode='r') as zip_io:
            labels_io = io.TextIOWrapper(zip_io.open("labels.txt"), encoding='utf-8')
            features_io = io.TextIOWrapper(zip_io.open("features.txt"), encoding='utf-8')
            uuids_io = io.TextIOWrapper(zip_io.open("uuids.txt"), encoding='utf-8')
            images_io = io.TextIOWrapper(zip_io.open("images.txt"), encoding='utf-8')

            while True:
                label = labels_io.readline().strip()
                feature = features_io.readline().strip()
                uuid_ = uuids_io.readline().strip()
                image_url = images_io.readline().strip()

                if label == '' or feature == '':
                    break

                batch.append(DataEntry(
                    label,
                    feature,
                    uuid_,
                    image_url
                ))
                # if len(batch) >= self.concurrency:
                if len(batch) >= self.concurrency:
                    yield batch
                    batch = []

            if batch:
                yield batch



import os

import click

from rebotics_sdk.providers import ProviderHTTPClientException


def task_create_processing_action_for_image(ctx, store_id, image_path):
    under_images_folder = os.path.join(os.getcwd(), 'images', image_path)
    cwd_filepath = os.path.join(os.getcwd(), image_path)
    if os.path.exists(under_images_folder):
        filepath = under_images_folder
    elif os.path.exists(cwd_filepath):
        filepath = cwd_filepath
    else:
        if ctx.verbose:
            click.echo('File does not exist {}'.format(image_path))
        return None

    with open(filepath, 'rb') as file_io:
        processing_input_dict = ctx.provider.processing_upload(store_id, file_io)
        if ctx.verbose:
            click.echo('File uploaded with id {}'.format(processing_input_dict['id']))

    processing_action_dict = ctx.provider.create_processing_action(
        store_id,
        [processing_input_dict['id']],
    )
    if ctx.verbose:
        click.echo('Processing action created with id {}'.format(processing_action_dict['id']))

    return processing_action_dict['id']


def upload_preview_task(d):
    ctx = d['ctx']
    images_path = d['images_path']
    upc = d['upc']
    delete = d['delete']

    if delete:
        try:
            response = ctx.provider.delete_product_previews(upc)
            click.echo('Deleted %s previews for product %s' % (response['deleted'], upc,), err=True)
        except ProviderHTTPClientException:
            click.echo('Product with UPC %s does not exists' % upc, err=True)

    for image_path in images_path:
        with open(image_path, 'rb') as image:
            try:
                ctx.provider.upload_product_preview(upc, image)
            except ProviderHTTPClientException as e:
                click.echo("xxx Failed to send image: %s" % image_path)
                click.echo(str(e), err=True)
            click.echo('<< Uploaded product preview for %s from %s' % (upc, image_path), err=True)


def task_download_processing_action(ctx, processing_action_id):
    try:
        processing_action_id = int(processing_action_id)
    except (ValueError, TypeError):
        processing_action_id = None

    if processing_action_id:
        if ctx.verbose:
            click.echo('Downloading processing action {}'.format(processing_action_id), err=True)
        try:
            return ctx.provider.processing_action_detail(processing_action_id)
        except ProviderHTTPClientException:
            return None


def task_recalculate_processing(d):
    ctx = d['ctx']
    provider = ctx.provider
    response = provider.recalculate(d['id'])
    if ctx.verbose:
        click.echo(response)
        click.echo(response['id'])
    return response


def task_cancel_processing(d):
    ctx = d['ctx']
    provider = ctx.provider
    response = provider.cancel(d['id'])
    if ctx.verbose:
        click.echo(response)
        click.echo(response['id'])
    return response


def task_requeue_processing(d):
    ctx = d['ctx']
    provider = ctx.provider
    response = provider.requeue(d['id'], d['requeue_type'])
    if ctx.verbose:
        click.echo(response)
        click.echo(response['id'])
    return response



import os
import pathlib

import click

from rebotics_sdk.cli.common import configure, shell, roles, set_token
from rebotics_sdk.cli.utils import read_saved_role, process_role, ReboticsCLIContext, app_dir, pass_rebotics_context
from rebotics_sdk.providers.fvm import FVMProvider
from .. import utils
from ..advanced import remote_loaders
from ..providers import ProviderHTTPClientException


@click.group()
@click.option('-f', '--format', default='table', type=click.Choice(['table', 'id', 'json']), help='Result rendering')
@click.option('-v', '--verbose', is_flag=True, help='Enables verbose mode')
@click.option('-c', '--config', type=click.Path(), default='fvm.json', help="Specify what config.json to use")
@click.option('-r', '--role', default=lambda: read_saved_role('fvm'), help="Key to specify what fvm to use")
@click.version_option()
@click.pass_context
def api(ctx, format, verbose, config, role):
    """
    Admin CLI tool to communicate with FVM API
    """
    process_role(ctx, role, 'fvm')
    ctx.obj = ReboticsCLIContext(
        role,
        format,
        verbose,
        os.path.join(app_dir, config),
        provider_class=FVMProvider
    )


api.add_command(shell, 'shell')
api.add_command(roles, 'roles')
api.add_command(configure, 'configure')
api.add_command(set_token, 'set_token')


@api.command(name='file')
@click.option('-f', '--name', required=True, help='Filename of the file', type=click.UNPROCESSED)
@pass_rebotics_context
def virtual_upload(ctx, name):
    """Create virtual upload"""
    if ctx.verbose:
        click.echo('Calling create virtual upload')
    result = ctx.provider.create_virtual_upload(
        name
    )
    if 'id' in result.keys():
        pk = result['id']
        with open(name, 'rb', ) as fio:
            click.echo('Uploading file...')
            remote_loaders.upload(destination=result['destination'], file=fio, filename=name)
            ctx.provider.finish(
                pk
            )
            click.echo("Successfully finished uploading")
    else:
        click.echo("Failed to call virtual upload")


@api.group()
def rcdb():
    pass


def _download_rcdb_locally(ctx, rcdb_file, target):
    rcdb_file = str(rcdb_file)

    if rcdb_file.isdigit():
        # download rcdb file by ID
        try:
            response = ctx.provider.get_rcdb_by_id(int(rcdb_file))
        except ProviderHTTPClientException as exc:
            raise click.ClickException(f"Failed to trigger by API with error: {exc}")
        # should we also save a response? probably no
        return _download_rcdb_locally(ctx, response['file']['file'], target)  # get the url
    elif utils.is_url(rcdb_file):
        # download rcdb file by URL
        filename = utils.get_filename_from_url(rcdb_file)
        local_filepath = pathlib.Path(target / filename)
        try:
            remote_loaders.download(rcdb_file, local_filepath)
        except Exception as exc:
            raise click.ClickException(f"Failed to download file by URL: {exc}")
        return _download_rcdb_locally(ctx, local_filepath, target)
    else:
        # assuming that it is a local file path
        local_filepath = pathlib.Path(rcdb_file)
        if not local_filepath.exists():
            raise click.ClickException("Local file is not loaded!")
        return local_filepath


@rcdb.command(name="unpack")
@click.argument('rcdb_file')
@click.option('-t', '--target', type=click.Path(), default='.')
@click.option('-w', '--with-images', is_flag=True)
@pass_rebotics_context
def rcdb_unpack(ctx, rcdb_file, target, with_images):
    rcdb_file = _download_rcdb_locally(ctx, rcdb_file, target)
    # create new packer and unpacker




import json
import os
import pathlib
import shutil
from functools import partial
from multiprocessing.pool import Pool

import click
import requests
from tqdm import tqdm

from rebotics_sdk.cli.common import shell, configure, roles, set_token
from .utils import ReboticsCLIContext, pass_rebotics_context, app_dir, read_saved_role, process_role
from ..advanced import remote_loaders
from ..providers import DatasetProvider, ProviderHTTPClientException
from ..providers.utils import get_id_list_from_ranges
from ..utils import download_file, mkdir_p, get_filename_from_url
from ..constants import DownloadImagesFormat


@click.group()
@click.option('-f', '--format', default='table', type=click.Choice(['table', 'id', 'json']))
@click.option('-v', '--verbose', is_flag=True, help='Enables verbose mode')
@click.option('-c', '--config', type=click.Path(), default='dataset.json', help="Specify what config.json to use")
@click.option('-r', '--role', default=lambda: read_saved_role('dataset'), help="Key to specify what dataset to use")
@click.version_option()
@click.pass_context
def api(ctx, format, verbose, config, role):
    """
    Dataset CLI tool to communicate with dataset API
    """
    process_role(ctx, role, 'dataset')
    ctx.obj = ReboticsCLIContext(
        role,
        format,
        verbose,
        os.path.join(app_dir, config),
        provider_class=DatasetProvider
    )


def save_image_task(d):
    download_file(d['url'], d['filepath'])
    click.echo('Downloaded jpeg image to {}'.format(d['filepath']))


@api.command()
@click.argument('meta_data_url', )
@click.option('-t', '--target', type=click.Path(), default='.')
@click.option('-c', '--concurrency', type=int, default=4)
@pass_rebotics_context
def setup_training_dir(ctx, meta_data_url, target, concurrency):
    """Set up training directory"""
    click.echo("Downloading meta data from {}".format(meta_data_url), err=True)
    response = requests.get(meta_data_url, timeout=remote_loaders.HTTP_TIMEOUTS)
    response.raise_for_status()

    meta_data = response.json()
    click.echo("Downloaded meta data from {}".format(meta_data_url), err=True)

    mkdir_p(target)
    annotations_path = os.path.join(target, 'Annotations')
    mkdir_p(annotations_path)
    images_path = os.path.join(target, 'JPEGImages')
    mkdir_p(images_path)
    image_sets_path = os.path.join(target, 'ImageSets')
    mkdir_p(image_sets_path)
    main_path = os.path.join(image_sets_path, 'Main')
    mkdir_p(main_path)

    for file_in_main in meta_data['main']:
        with open(os.path.join(main_path, file_in_main['filename']), 'w') as fout:
            fout.write(file_in_main['value'])

    for annotation_file in meta_data['annotations']:
        with open(os.path.join(annotations_path, annotation_file['filename']), 'w') as fout:
            fout.write(annotation_file['value'])

    click.echo("Saved annotation files. Fetching images...", err=True)

    p = Pool(concurrency)
    jpeg_images = [
        {
            'url': i['url'],
            'filepath': os.path.join(images_path, i['image_name'])
        } for i in meta_data['jpeg_images']
    ]
    p.map(save_image_task, jpeg_images)


def task_upload_feature_vector_batch(d):
    ctx, data, batch_index = d
    try:
        return ctx.provider.save_feature_vector_batch(data)
    except ProviderHTTPClientException as exc:
        return exc


@api.command()
@click.argument('ref_upc', type=click.File())
@click.argument('ref_features', type=click.File())
@click.argument('ref_threshold', type=click.File())
@click.option('-r', '--retailer', prompt=True, )
@click.option('-f', '--facenet', prompt=True, )
@click.option('-s', '--batch-size', default=100)
@click.option('-c', '--concurrency', type=int, default=4)
@pass_rebotics_context
def upload_ref(ctx, ref_upc, ref_features, ref_threshold, retailer, facenet, batch_size, concurrency):
    """Uploads reference UPC and reference features for retailer and facenet to dataset"""
    click.echo('Starting uploading to {provider.host} for {retailer} and {facenet}...'.format(
        provider=ctx.provider, retailer=retailer, facenet=facenet
    ), err=True)

    if ctx.verbose:
        click.echo('Generating batches')

    data_to_post = [
        ctx.provider.construct_reference_entry(
            upc=upc_code, feature=feature_vector, threshold=threshold,
            retailer=retailer, facenet=facenet,
        )
        for upc_code, feature_vector, threshold in zip(ref_upc, ref_features, ref_threshold)
    ]

    batches = []

    for batch_index in range(0, len(data_to_post), batch_size):
        batches.append(data_to_post[batch_index:batch_index + batch_size])

    if ctx.verbose:
        click.echo('Created {} batches'.format(len(batches)))

    pool = Pool(concurrency)

    result = []
    for res in tqdm(
            pool.imap_unordered(task_upload_feature_vector_batch, [(ctx, d, i) for i, d in enumerate(batches)]),
            total=len(batches)
    ):
        result.append(res)

    if ctx.verbose:
        click.echo(result)

    click.echo('Done uploading to {provider.host} for {retailer} and {facenet}!'.format(
        provider=ctx.provider, retailer=retailer, facenet=facenet
    ))


@api.command()
@click.option('-r', '--retailer', prompt=True, )
@click.option('-f', '--facenet', prompt=True, )
@click.argument('output', type=click.Path(), default='.')
@pass_rebotics_context
def download_ref_backup(ctx, retailer, facenet, output):
    """Downloads reference backup for retailer and facenet"""
    filename, raw = ctx.provider.download_reference_database(retailer, facenet)

    if os.path.isdir(output):
        mkdir_p(output)
        output = os.path.join(output, filename)
    elif os.path.isfile(output):
        mkdir_p(os.path.dirname(output))
    else:
        raise OSError('Could not write to path that does not')

    if ctx.verbose:
        click.echo(output)

    with open(output, 'wb') as out_file:
        shutil.copyfileobj(raw, out_file)


def task_download_and_set_training_data(item_id, target, format, ctx):
    try:
        response = ctx.provider.get_detection_image_by_id(item_id)
    except:
        click.echo("Failed to find image with id={}".format(item_id))
        return
    # download images and download this json
    if format in (DownloadImagesFormat.FULL, DownloadImagesFormat.IMAGES):
        image_url = response['image']
        image_file = target / "{}.{ext}".format(
            item_id,
            ext=get_filename_from_url(image_url).split(".")[-1]
        )
        try:
            remote_loaders.download(image_url, str(image_file))
        except:
            click.echo("{} failed to download image from {}".format(item_id, image_url))
    if format in (DownloadImagesFormat.FULL, DownloadImagesFormat.ANNOTATIONS):
        json_file = target / "{}.json".format(item_id)
        with open(str(json_file), 'w') as fio:
            json.dump(response, fio, indent=4)


@api.command()
@click.argument("image_ids", type=click.STRING)
@click.option('-t', '--target', type=click.Path(), default='.')
@click.option('-c', '--concurrency', type=int, default=4)
@click.option('-f', '--format',
              type=click.Choice(DownloadImagesFormat.OPTIONS, case_sensitive=False),
              default=DownloadImagesFormat.FULL,)
@pass_rebotics_context
def download_images(ctx, image_ids, target, concurrency, format):
    """Download images and/or annotations json."""
    # parse image_ids as range or as a ids
    id_list = get_id_list_from_ranges(image_ids)
    if ctx.verbose:
        click.echo(id_list)

    if not id_list:
        click.echo("Invalid ID list is supplied, failed to extract ids")

    mkdir_p(str(target))

    with Pool(min(len(id_list), concurrency)) as pool:
        pool.map(
            partial(task_download_and_set_training_data,
                    target=pathlib.Path(target),
                    format=format,
                    ctx=ctx,),
            id_list
        )
    pool.join()
    pool.close()


api.add_command(shell, 'shell')
api.add_command(roles, 'roles')
api.add_command(configure, 'configure')
api.add_command(set_token, 'set_token')



import pytest

from click.testing import CliRunner


@pytest.fixture(scope="function")
def runner():
    return CliRunner()


@pytest.fixture(scope="function")
def mocker_patch_provider(mocker):
    from unittest.mock import Mock
    from rebotics_sdk.providers.hawkeye import HawkeyeProvider
    process_role_mock = mocker.patch("rebotics_sdk.cli.utils.process_role")
    process_role_mock.return_value = None
    provider_mock = Mock(spec=HawkeyeProvider)
    mocker.patch('rebotics_sdk.cli.utils.ReboticsCLIContext.provider', new=provider_mock)
    return provider_mock
